// Cytosim was created by Francois Nedelec. Copyright 2007-2017 EMBL.
/**
 @file
 @brief  Dimensionality of the simulation
 */

#ifndef DIM

/**
 DIM defines the number of dimensions in space:
 it must be 1, 2 or 3
 */
#define DIM 2

#endif
